#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Jun 18 14:44:33 2019

@author: jonasg
"""

var_groups = {'EBASMC' : [
                ['scatc550aer', 'scatc550dryaer','absc550aer'],
                ['conco3'],
                ['concpm10', 'concpm25']]}

